#include "matrix.hh"

bool is_zero(const qq_matrix& m) {
  for(qq_matrix::const_iterator1 line=m.begin1();
      line != m.end1(); line++) {
    for(qq_matrix::const_iterator1::dual_iterator_type pos=line.begin();
	pos != line.end(); pos++) {
      if (! (*pos).isZero()) return false;
    }
  }
  return true;
}

std::ostream& operator<<(std::ostream& o, const qq_vector& v) {
  o << "[ ";
  for(mat_index i=0; i<v.size(); i++) {
    if (i > 0) o << ", ";
    o << v[i];
  }
  o << " ]";
  return o;
}

std::ostream& operator<<(std::ostream& o, const qq_matrix& m) {
  for(mat_index i=0; i<m.size1(); i++) {
    o << "[ ";
    for(mat_index j=0; j<m.size2(); j++) {
      if (j > 0) {
	o << ", ";
      }
      o << m(i, j);
    }
    o << " ]" << std::endl;
  }
  return o;
}

std::ostream& operator<<(std::ostream& o, const ff_vector& v) {
  o << "[ ";
  for(mat_index i=0; i<v.size(); i++) {
    if (i > 0) o << ", ";
    o << v[i];
  }
  o << " ]";
  return o;
}

std::ostream& operator<<(std::ostream& o, const ff_matrix& m) {
  for(mat_index i=0; i<m.size1(); i++) {
    o << "[ ";
    for(mat_index j=0; j<m.size2(); j++) {
      if (j > 0) {
	o << ", ";
      }
      o << m(i, j);
    }
    o << " ]" << std::endl;
  }
  return o;
}

mat_index echelonize(qq_matrix& m, std::vector<mat_index>* pivot_columns) {
  mat_index i=0;
  for(mat_index j=0; j<m.size2(); j++) {
    matrix_column<qq_matrix> col(m, j);
    {
      mat_index i2;
      matrix_column<qq_matrix>::const_iterator pos;
      for(pos=col.find(i); pos != col.end() && (*pos).isZero(); pos++) {}
      if (pos == col.end()) continue;
      i2 = pos.index();
      if (i != i2) row(m, i).swap(row(m, i2));
    }
    matrix_row<qq_matrix> rmi(m, i);
    QQ coeff(m(i, j));
    rmi /= coeff;
    for(matrix_column<qq_matrix>::const_iterator pos=col.begin();
	pos != col.end(); pos++) {
      mat_index i2 = pos.index();
      if (i2==i || (*pos).isZero()) continue;
      row(m, i2) -= m(i2, j) * rmi;
    }
    if (pivot_columns) (*pivot_columns)[i]=j;
    i++;
  }
  if (pivot_columns) (*pivot_columns)[i]=m.size2();
  return i;
}

qq_matrix right_kernel(const qq_matrix& m_src) {
  qq_matrix m = m_src;
#ifdef EXPENSIVE_ASSERT
  qq_matrix ker = echelonize_and_right_kernel(m);
  assert(is_zero(prod(m_src,trans(ker))));
  return ker;
#else
  return echelonize_and_right_kernel(m);
#endif
}

qq_matrix echelonize_and_right_kernel(qq_matrix& m) {
  std::vector<mat_index> pivot_columns(m.size1()+1);
  mat_index rank = echelonize(m, &pivot_columns);
  qq_matrix ker(m.size2()-rank,m.size2());

  mat_index pivot_i=0, ker_i=0;
  for(mat_index j=0; j<m.size2(); j++) {
    if (j == pivot_columns[pivot_i]) {
      pivot_i++;
      continue;
    }
    ker(ker_i, j) = 1;
    matrix_column<qq_matrix> col(m, j);
    for(matrix_column<qq_matrix>::const_iterator pos=col.begin();
	pos != col.end(); pos++) {
      if ((*pos).isZero()) continue;
      ker(ker_i, pivot_columns[pos.index()]) = -(*pos);
    }
    ker_i++;
  }
  assert(ker_i == ker.size1());
  return ker;
}

generator_system::generator_system(mat_index nr_vectors, mat_index dimension):
  empty_flag(false),
  f0(dimension),
  fi(nr_vectors, dimension) {
  f0.clear();
  fi.clear();
}

generator_system::generator_system(mat_index dimension):
  empty_flag(false),
  f0(dimension),
  fi(dimension, dimension) {
  f0.clear();
  fi.clear();
  for(mat_index i=0; i<dimension; i++) {
    fi(i, i) = 1;
  }
}

generator_system::generator_system(const qq_vector& f0_p, const std::vector<qq_vector>& fi_p) :
  empty_flag(false),
  f0(f0_p),
  fi(fi_p.size(), f0_p.size()) {
  for(mat_index i=0; i<fi_p.size(); i++) {
    assert(fi.size2() == fi_p[i].size());
    row(fi, i) = fi_p[i];
  }
}

void generator_system::print(std::ostream& o) const {
  if (empty_flag) {
    o << "<empty>" << std::endl;
  } else {
    o << "Origin" << std::endl << f0 << std::endl;
    o << "Generator rows" << std::endl << fi;
  }
}

void generator_system::constrain(const qq_matrix& constraints) {
  assert(constraints.size2() == fi.size2()+1);
  if (empty_flag) return;

  qq_matrix gens(1+dimension(), 1+ambient_dimension());
  matrix_vector_slice<qq_matrix>(gens, slice(fi.size1(), 0, ambient_dimension()), slice(0, 1, ambient_dimension())) = f0;
  gens(fi.size1(), ambient_dimension()) = 1;
  subrange(gens, 0, fi.size1(), 0, fi.size2()) = fi;

  // std::cout << "gens" << std::endl << gens;

  qq_matrix ker = right_kernel(prod(constraints, trans(gens)));
  // std::cout << "ker" << std::endl << ker;

  if (ker.size1() == 0) {
    empty_flag = true;
    return;
  }

  for(mat_index i=0; i<ker.size1()-1; i++) {
    assert(QQ(ker(i, ker.size2()-1)).isZero());
  }

  if (ker(ker.size1()-1, ker.size2()-1) == 0) {
    empty_flag = true;
    return;
  }

  assert(ker(ker.size1()-1, ker.size2()-1) == 1);

  f0 += prod(matrix_vector_slice<qq_matrix>(ker, slice(ker.size1()-1, 0, ker.size2()-1), slice(0, 1, ker.size2()-1)), fi);
  fi = prod(subrange(ker, 0, ker.size1()-1, 0, ker.size2()-1), fi);
}

qq_vector generator_system::at(const qq_vector& y) const
{
  assert(y.size() == fi.size1());
  return f0 + prod(y, fi);
}

QQ rationalize(FF x, FF epsilon, int depth) {
  if (x < 0) return -rationalize(-x, epsilon, depth);
  ZZ integral_part(x + 0.5);
  if (depth==0) return QQ(integral_part);
  FF delta = x-integral_part.get_d();
  if (fabs(delta) < epsilon) return QQ(integral_part);
  return QQ(integral_part) + rationalize(FF(1.0) / delta, epsilon, depth-1).inverse();
}

sum_of_squares sos_decompose(qq_matrix m) throw(Is_not_sdp) {
  assert(m.size1() == m.size2());
  sum_of_squares sos;
  for(unsigned i=0; i<m.size1(); i++) {
    QQ pivot = m(i, i);
    if (pivot < 0) throw Is_not_sdp();
    if (pivot == 0) {
      for(unsigned j=i+1; j<m.size1(); j++) {
	if (m(i, j) != 0) throw Is_not_sdp();
      }
    } else {
      qq_vector v = row(m, i) / pivot;
      sos.push_back(std::pair<QQ, qq_vector>(pivot, v));
      qq_matrix h(1, v.size());
      row(h, 0) = v;
      m -= pivot * prod(trans(h),h);
    }
  }
  return sos;
}

qq_matrix sos_recompose(unsigned dim, const sum_of_squares& sos) {
  qq_matrix m(dim, dim);
  for(sum_of_squares::const_iterator k = sos.begin();
      k != sos.end(); k++) {
    const QQ& c = k->first;
    const qq_vector& v = k->second;
    qq_matrix h(1, v.size());
      row(h, 0) = v;
      m += c * prod(trans(h),h);
  }
  return m;
}

bool is_sdp(qq_matrix m) {
  assert(m.size1() == m.size2());
  for(unsigned i=0; i<m.size1(); i++) {
    QQ pivot = m(i, i);
    if (pivot < 0) return false;
    if (pivot == 0) {
      for(unsigned j=i+1; j<m.size1(); j++) {
	if (m(i, j) != 0) return false;
      }
    } else {
      qq_vector v = row(m, i) / pivot;
      qq_matrix h(1, v.size());
      row(h, 0) = v;
      m -= pivot * prod(trans(h),h);
    }
  }
  return true;
}

bool is_dp(qq_matrix m) {
  assert(m.size1() == m.size2());
  for(unsigned i=0; i<m.size1(); i++) {
    QQ pivot = m(i, i);
    if (pivot <= 0) return false;
    qq_vector v = row(m, i) / pivot;
    qq_matrix h(1, v.size());
    row(h, 0) = v;
    m -= pivot * prod(trans(h),h);
  }
  return true;
}


ff_vector qq_to_ff_vector(const qq_vector& v) {
  ff_vector ret(v.size());
  ret.clear();
  for(qq_vector::const_iterator it = v.begin();
      it != v.end(); it++) {
    if ((*it).isZero()) continue;
    ret[it.index()] = (*it).get_d();
  }
  return ret;
}

ff_matrix qq_to_ff_matrix(const qq_matrix& m) {
  ff_matrix ret(m.size1(), m.size2());
  ret.clear();
  for(unsigned i=0; i<m.size1(); i++) {
    matrix_row<const qq_matrix> rowm(m, i);
    for(matrix_row<const qq_matrix>::const_iterator pos = rowm.begin();
	pos != rowm.end(); pos++) {
      ret(i, pos.index()) = (*pos).get_d(); 
    }
  }
  return ret;
}

ff_vector generator_system::at(const ff_vector& y) const
{
  assert(y.size() == fi.size1());
  return qq_to_ff_vector(f0) + prod(y, qq_to_ff_matrix(fi));
}

extern "C" void dgels_(const char *trans, const int *m, const int *n, const int *nrhs, double *a, const int *lda, double *b, const int *ldb, double *work, const int *lwork, int *info);

ff_vector generator_system::project(const ff_vector& v) const {
  assert(v.size() == f0.size());
  assert(v.size() == fi.size2());
  if (fi.size1()==0) return ff_vector();
  assert(fi.size1() <= fi.size2());

  const int nrows=fi.size1(), ncols=fi.size2(), nrhs=1, lda=ncols, ldb=ncols;
  typedef double row[ncols];
  row *a = new row[nrows];
  double b[ldb];
  for(int i=0; i<nrows; i++) {
    for(int j=0; j<ncols; j++) {
      a[i][j] = fi(i, j).get_d();
    }
  }
  for(int j=0; j<ncols; j++) {
    b[j] = v(j) - f0(j).get_d();
  }

  int lwork = -1, info;
  double temp;
  dgels_("N", &ncols, &nrows, &nrhs, &a[0][0], &lda, b, &ldb, &temp, &lwork, &info);
  assert(info == 0);

  lwork = (int) temp;
  double work[lwork];
  dgels_("N", &ncols, &nrows, &nrhs, &a[0][0], &lda, b, &ldb, work, &lwork, &info);
  assert(info == 0);

  delete[] a;

  ff_vector ret(nrows);
  for(int i=0; i<nrows; i++) {
    ret(i) = b[i];
  }

  double s = 0.0;
  for(int i=nrows; i<ldb; i++) {
    s += b[i]*b[i];
  }
  std::cout << "least square distance: " << sqrt(s) << std::endl;
  return ret;
}

qq_matrix build_from_rows(const std::vector<qq_vector>& rows) {
  assert (rows.size() > 0);
  qq_matrix m(rows.size(), rows[0].size());
  for(unsigned i=0; i<rows.size(); i++) {
    row(m, i) = rows[i];
  }
  return m;
}
